"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var brightscript_parser_1 = require("brightscript-parser");
var trimRight = require("trim-right");
var BrightScriptFormatter = /** @class */ (function () {
    function BrightScriptFormatter() {
    }
    /**
     * Format the given input.
     * @param inputText the text to format
     * @param formattingOptions options specifying formatting preferences
     */
    BrightScriptFormatter.prototype.format = function (inputText, formattingOptions) {
        var options = this.normalizeOptions(formattingOptions);
        var lexer = new brightscript_parser_1.BrightScriptLexer();
        var tokens = lexer.tokenize(inputText);
        //force all composite keywords to have 0 or 1 spaces in between, but no more than 1
        tokens = this.normalizeCompositeKeywords(tokens);
        if (options.compositeKeywords) {
            tokens = this.formatCompositeKeywords(tokens, options);
        }
        tokens = this.formatKeywordCase(tokens, options);
        if (options.removeTrailingWhiteSpace) {
            tokens = this.formatTrailingWhiteSpace(tokens, options);
        }
        if (options.formatInteriorWhitespace) {
            tokens = this.formatInteriorWhitespace(tokens, options);
        }
        //dedupe side-by-side whitespace tokens
        this.dedupeWhitespace(tokens);
        if (options.formatIndent) {
            tokens = this.formatIndentation(tokens, options);
        }
        //join all tokens back together into a single string
        var outputText = '';
        for (var _i = 0, tokens_1 = tokens; _i < tokens_1.length; _i++) {
            var token = tokens_1[_i];
            outputText += token.value;
        }
        return outputText;
    };
    /**
     * Remove all whitespace in the composite keyword tokens with a single space
     * @param tokens
     */
    BrightScriptFormatter.prototype.normalizeCompositeKeywords = function (tokens) {
        var indexOffset = 0;
        for (var _i = 0, tokens_2 = tokens; _i < tokens_2.length; _i++) {
            var token = tokens_2[_i];
            token.startIndex += indexOffset;
            //is this a composite token
            if (brightscript_parser_1.CompositeKeywordTokenTypes.indexOf(token.tokenType) > -1) {
                var value = token.value;
                //remove all whitespace with a single space
                token.value.replace(/s+/g, ' ');
                var indexDifference = value.length - token.value.length;
                indexOffset -= indexDifference;
            }
        }
        return tokens;
    };
    BrightScriptFormatter.prototype.dedupeWhitespace = function (tokens) {
        for (var i = 0; i < tokens.length; i++) {
            var currentToken = tokens[i];
            var nextToken = tokens[i + 1] ? tokens[i + 1] : { tokenType: undefined, value: '' };
            if (currentToken.tokenType === brightscript_parser_1.TokenType.whitespace && nextToken.tokenType === brightscript_parser_1.TokenType.whitespace) {
                currentToken.value += nextToken.value;
                tokens.splice(i + 1, 1);
                //decrement the counter so we process this token again so it can absorb more whitespace tokens
                i--;
            }
        }
    };
    BrightScriptFormatter.prototype.formatCompositeKeywords = function (tokens, options) {
        var indexOffset = 0;
        for (var _i = 0, tokens_3 = tokens; _i < tokens_3.length; _i++) {
            var token = tokens_3[_i];
            token.startIndex += indexOffset;
            //is this a composite token
            if (brightscript_parser_1.CompositeKeywordTokenTypes.indexOf(token.tokenType) > -1) {
                var parts = this.getCompositeKeywordParts(token);
                var tokenValue = token.value;
                if (options.compositeKeywords === 'combine') {
                    token.value = parts[0] + parts[1];
                }
                else if (options.compositeKeywords === 'split') {
                    // if(options.compositeKeywords === 'split'){
                    token.value = parts[0] + ' ' + parts[1];
                }
                else {
                    //do nothing
                }
                var offsetDifference = token.value.length - tokenValue.length;
                indexOffset += offsetDifference;
            }
        }
        return tokens;
    };
    BrightScriptFormatter.prototype.getCompositeKeywordParts = function (token) {
        var lowerValue = token.value.toLowerCase();
        //split the parts of the token, but retain their case
        if (lowerValue.indexOf('end') === 0) {
            return [token.value.substring(0, 3), token.value.substring(3).trim()];
        }
        else if (lowerValue.indexOf('#else') === 0) {
            return [token.value.substring(0, 5), token.value.substring(5).trim()];
        }
        else {
            // if (lowerValue.indexOf('exit') === 0 || lowerValue.indexOf('else') === 0) {
            return [token.value.substring(0, 4), token.value.substring(4).trim()];
        }
    };
    /**
     * Determine if the token is a type keyword (meaing preceeded by `as` token)
     * @param token
     */
    BrightScriptFormatter.prototype.isType = function (tokens, token) {
        var previousToken = this.getPreviousNonWhitespaceToken(tokens, tokens.indexOf(token));
        if (previousToken && previousToken.value.toLowerCase() === 'as') {
            return true;
        }
        else {
            return false;
        }
    };
    BrightScriptFormatter.prototype.formatKeywordCase = function (tokens, options) {
        for (var _i = 0, tokens_4 = tokens; _i < tokens_4.length; _i++) {
            var token = tokens_4[_i];
            //if this token is a keyword
            if (brightscript_parser_1.KeywordTokenTypes.indexOf(token.tokenType) > -1) {
                //a token is a type if it's preceeded by an `as` token
                var isType = this.isType(tokens, token);
                var keywordCase = void 0;
                if (isType) {
                    keywordCase = options.typeCase;
                }
                else {
                    keywordCase = options.keywordCase;
                    //if this is an overridable keyword, use that override instead
                    if (options.keywordCaseOverride && options.keywordCaseOverride[token.tokenType] !== undefined) {
                        keywordCase = options.keywordCaseOverride[token.tokenType];
                    }
                }
                switch (keywordCase) {
                    case 'lower':
                        token.value = token.value.toLowerCase();
                        break;
                    case 'upper':
                        token.value = token.value.toUpperCase();
                        break;
                    case 'title':
                        var lowerValue = token.value.toLowerCase();
                        if (brightscript_parser_1.CompositeKeywordTokenTypes.indexOf(token.tokenType) === -1) {
                            token.value =
                                token.value.substring(0, 1).toUpperCase() +
                                    token.value.substring(1).toLowerCase();
                        }
                        else {
                            var spaceCharCount = (lowerValue.match(/\s+/) || []).length;
                            var firstWordLength = 0;
                            if (lowerValue.indexOf('end') === 0) {
                                firstWordLength = 3;
                            }
                            else {
                                //if (lowerValue.indexOf('exit') > -1 || lowerValue.indexOf('else') > -1)
                                firstWordLength = 4;
                            }
                            token.value =
                                //first character
                                token.value.substring(0, 1).toUpperCase() +
                                    //rest of first word
                                    token.value.substring(1, firstWordLength).toLowerCase() +
                                    //add back the whitespace
                                    token.value.substring(firstWordLength, firstWordLength + spaceCharCount) +
                                    //first character of second word
                                    token.value
                                        .substring(firstWordLength + spaceCharCount, firstWordLength + spaceCharCount + 1)
                                        .toUpperCase() +
                                    //rest of second word
                                    token.value
                                        .substring(firstWordLength + spaceCharCount + 1)
                                        .toLowerCase();
                        }
                    case 'original':
                    case null:
                    case undefined:
                    default:
                        //do nothing
                        break;
                }
            }
        }
        return tokens;
    };
    BrightScriptFormatter.prototype.formatIndentation = function (tokens, options) {
        var indentTokens = [
            brightscript_parser_1.TokenType.sub,
            brightscript_parser_1.TokenType.for,
            brightscript_parser_1.TokenType.function,
            brightscript_parser_1.TokenType.if,
            brightscript_parser_1.TokenType.openCurlyBraceSymbol,
            brightscript_parser_1.TokenType.openSquareBraceSymbol,
            brightscript_parser_1.TokenType.while,
            brightscript_parser_1.TokenType.condIf
        ];
        var outdentTokens = [
            brightscript_parser_1.TokenType.closeCurlyBraceSymbol,
            brightscript_parser_1.TokenType.closeSquareBraceSymbol,
            brightscript_parser_1.TokenType.endFunction,
            brightscript_parser_1.TokenType.endIf,
            brightscript_parser_1.TokenType.endSub,
            brightscript_parser_1.TokenType.endWhile,
            brightscript_parser_1.TokenType.endFor,
            brightscript_parser_1.TokenType.next,
            brightscript_parser_1.TokenType.condEndIf
        ];
        var interumTokens = [
            brightscript_parser_1.TokenType.else,
            brightscript_parser_1.TokenType.elseIf,
            brightscript_parser_1.TokenType.condElse,
            brightscript_parser_1.TokenType.condElseIf
        ];
        var tabCount = 0;
        var nextLineStartTokenIndex = 0;
        //the list of output tokens
        var outputTokens = [];
        //set the loop to run for a max of double the number of tokens we found so we don't end up with an infinite loop
        outer: for (var outerLoopCounter = 0; outerLoopCounter <= tokens.length * 2; outerLoopCounter++) {
            var lineObj = this.getLineTokens(nextLineStartTokenIndex, tokens);
            nextLineStartTokenIndex = lineObj.stopIndex + 1;
            var lineTokens = lineObj.tokens;
            var thisTabCount = tabCount;
            var foundIndentorThisLine = false;
            //if this is a single-line if statement, do nothing with indentation
            if (this.isSingleLineIfStatement(lineTokens, tokens)) {
                // //if this line has a return statement, outdent
                // if (this.tokenIndexOf(TokenType.return, lineTokens) > -1) {
                //     tabCount--;
                // } else {
                //     //do nothing with single-line if statement indentation
                // }
            }
            else {
                inner: for (var i = 0; i < lineTokens.length; i++) {
                    var token = lineTokens[i];
                    var previousNonWhitespaceToken = this.getPreviousNonWhitespaceToken(lineTokens, i);
                    //if this is an indentor token,
                    if (indentTokens.indexOf(token.tokenType) > -1) {
                        //skip indent for 'function'|'sub' used as type (preceeded by `as` keyword)
                        if ((token.tokenType === brightscript_parser_1.TokenType.function || token.tokenType === brightscript_parser_1.TokenType.sub) &&
                            //the previous token will be whitespace, so verify that previousPrevious is 'as'
                            previousNonWhitespaceToken && previousNonWhitespaceToken.value.toLowerCase() === 'as') {
                            continue inner;
                        }
                        tabCount++;
                        foundIndentorThisLine = true;
                        //this is an outdentor token
                    }
                    else if (outdentTokens.indexOf(token.tokenType) > -1) {
                        tabCount--;
                        if (foundIndentorThisLine === false) {
                            thisTabCount--;
                        }
                        //this is an interum token
                    }
                    else if (interumTokens.indexOf(token.tokenType) > -1) {
                        //these need outdented, but don't change the tabCount
                        thisTabCount--;
                    }
                    //  else if (token.tokenType === TokenType.return && foundIndentorThisLine) {
                    //     //a return statement on the same line as an indentor means we don't want to indent
                    //     tabCount--;
                    // }
                }
            }
            //if the tab counts are less than zero, something is wrong. However, at least try to do formatting as best we can by resetting to 0
            thisTabCount = thisTabCount < 0 ? 0 : thisTabCount;
            tabCount = tabCount < 0 ? 0 : tabCount;
            var leadingWhitespace = void 0;
            if (options.indentStyle === 'spaces') {
                var indentSpaceCount = options.indentSpaceCount
                    ? options.indentSpaceCount
                    : BrightScriptFormatter.DEFAULT_INDENT_SPACE_COUNT;
                var spaceCount = thisTabCount * indentSpaceCount;
                leadingWhitespace = Array(spaceCount + 1).join(' ');
            }
            else {
                leadingWhitespace = Array(thisTabCount + 1).join('\t');
            }
            //create a whitespace token if there isn't one
            if (lineTokens[0] && lineTokens[0].tokenType !== brightscript_parser_1.TokenType.whitespace) {
                lineTokens.unshift({
                    startIndex: -1,
                    tokenType: brightscript_parser_1.TokenType.whitespace,
                    value: ''
                });
            }
            //replace the whitespace with the formatted whitespace
            lineTokens[0].value = leadingWhitespace;
            //add this list of tokens
            outputTokens.push.apply(outputTokens, lineTokens);
            //if we have found the end of file
            if (lineTokens[lineTokens.length - 1].tokenType === brightscript_parser_1.TokenType.END_OF_FILE) {
                break outer;
            }
            /* istanbul ignore next */
            if (outerLoopCounter === tokens.length * 2) {
                throw new Error('Something went terribly wrong');
            }
        }
        return outputTokens;
    };
    /**
     * Force all whitespace between tokens to be exactly 1 space wide
     */
    BrightScriptFormatter.prototype.formatInteriorWhitespace = function (tokens, options) {
        var addBoth = [
            //assignments
            brightscript_parser_1.TokenType.equalSymbol,
            brightscript_parser_1.TokenType.additionAssignmentSymbol,
            brightscript_parser_1.TokenType.subtractionAssignmentSymbol,
            brightscript_parser_1.TokenType.multiplicationAssignmentSymbol,
            brightscript_parser_1.TokenType.divisionAssignmentSymbol,
            brightscript_parser_1.TokenType.integerDivisionAssignmentSymbol,
            brightscript_parser_1.TokenType.lessThanLessThanEqualSymbol,
            brightscript_parser_1.TokenType.greaterThanGreaterThanEqualSymbol,
            //operators
            brightscript_parser_1.TokenType.plusSymbol,
            brightscript_parser_1.TokenType.minusSymbol,
            brightscript_parser_1.TokenType.asteriskSymbol,
            brightscript_parser_1.TokenType.forwardSlashSymbol,
            brightscript_parser_1.TokenType.backSlashSymbol,
            brightscript_parser_1.TokenType.carotSymbol,
            brightscript_parser_1.TokenType.notEqual,
            brightscript_parser_1.TokenType.lessThanOrEqual,
            brightscript_parser_1.TokenType.greaterThanOrEqual,
            brightscript_parser_1.TokenType.greaterThanSymbol,
            brightscript_parser_1.TokenType.lessThanSymbol,
        ];
        var addLeft = addBoth.concat([
            brightscript_parser_1.TokenType.closeCurlyBraceSymbol
        ]);
        var addRight = addBoth.concat([
            brightscript_parser_1.TokenType.openCurlyBraceSymbol,
            brightscript_parser_1.TokenType.commaSymbol,
            brightscript_parser_1.TokenType.colonSymbol
        ]);
        var removeBoth = [];
        var removeLeft = removeBoth.concat([
            brightscript_parser_1.TokenType.closeSquareBraceSymbol,
            brightscript_parser_1.TokenType.closeParenSymbol
        ]);
        var removeRight = removeBoth.concat([
            brightscript_parser_1.TokenType.openSquareBraceSymbol,
            brightscript_parser_1.TokenType.openParenSymbol
        ]);
        var isPastFirstTokenOfLine = false;
        for (var i = 0; i < tokens.length; i++) {
            var token = tokens[i];
            var nextTokenType = (tokens[i + 1] ? tokens[i + 1].tokenType : undefined);
            var previousTokenType = (tokens[i - 1] ? tokens[i - 1].tokenType : undefined);
            //reset token indicator on newline
            if (token.tokenType === brightscript_parser_1.TokenType.newline) {
                isPastFirstTokenOfLine = false;
                continue;
            }
            //skip past leading whitespace
            if (token.tokenType === brightscript_parser_1.TokenType.whitespace && isPastFirstTokenOfLine === false) {
                continue;
            }
            isPastFirstTokenOfLine = true;
            //force token to be exactly 1 space
            if (token.tokenType === brightscript_parser_1.TokenType.whitespace) {
                token.value = ' ';
            }
            //pad any of these token types with a space to the right
            if (addRight.indexOf(token.tokenType) > -1) {
                //special case: we want the negative sign to be directly beside a numeric, in certain cases.
                //we can't handle every case, but we can get close
                if (this.looksLikeNegativeNumericLiteral(tokens, i)) {
                    //throw out the space to the right of the minus symbol if present
                    if (i + 1 < tokens.length && tokens[i + 1].tokenType === brightscript_parser_1.TokenType.whitespace) {
                        this.removeWhitespace(tokens, i + 1);
                    }
                    //ensure a space token to the right, only if we have more tokens to the right available
                }
                else if ([brightscript_parser_1.TokenType.whitespace, brightscript_parser_1.TokenType.newline, brightscript_parser_1.TokenType.END_OF_FILE].indexOf(nextTokenType) === -1) {
                    //don't add whitespace if the next token is the newline
                    tokens.splice(i + 1, 0, {
                        startIndex: -1,
                        tokenType: brightscript_parser_1.TokenType.whitespace,
                        value: ' '
                    });
                }
            }
            //pad any of these tokens with a space to the left
            if (addLeft.indexOf(token.tokenType) > -1) {
                //ensure a space token to the left
                if (previousTokenType && previousTokenType !== brightscript_parser_1.TokenType.whitespace) {
                    tokens.splice(i, 0, {
                        startIndex: -1,
                        tokenType: brightscript_parser_1.TokenType.whitespace,
                        value: ' '
                    });
                    //increment i by 1 since we added a token
                    i++;
                }
            }
            //remove any space tokens on the right
            if (removeRight.indexOf(token.tokenType) > -1) {
                if (nextTokenType === brightscript_parser_1.TokenType.whitespace) {
                    //remove the next token, which is the whitespace token
                    tokens.splice(i + 1, 1);
                }
            }
            //remove any space tokens on the left
            if (removeLeft.indexOf(token.tokenType) > -1) {
                if (previousTokenType === brightscript_parser_1.TokenType.whitespace) {
                    //remove the previous token, which is the whitespace token
                    tokens.splice(i - 1, 1);
                    //backtrack the index since we just shifted the array
                    i--;
                }
            }
        }
        //handle special cases
        for (var i = 0; i < tokens.length; i++) {
            var token = tokens[i];
            var nextNonWhitespaceToken = this.getNextNonWhitespaceToken(tokens, i);
            //space to left of function parens?
            {
                var parenToken = void 0;
                //look for anonymous functions
                if (token.tokenType === brightscript_parser_1.TokenType.function && nextNonWhitespaceToken.tokenType === brightscript_parser_1.TokenType.openParenSymbol) {
                    parenToken = nextNonWhitespaceToken;
                    //look for named functions
                }
                else if (token.tokenType === brightscript_parser_1.TokenType.function && nextNonWhitespaceToken.tokenType === brightscript_parser_1.TokenType.identifier) {
                    //get the next non-whitespace token, which SHOULD be the paren
                    var parenCandidate = this.getNextNonWhitespaceToken(tokens, tokens.indexOf(nextNonWhitespaceToken));
                    if (parenCandidate.tokenType === brightscript_parser_1.TokenType.openParenSymbol) {
                        parenToken = parenCandidate;
                    }
                }
                //if we found the paren token, handle spacing
                if (parenToken) {
                    //walk backwards, removing any whitespace tokens found
                    this.removeWhitespaceTokensBackwards(tokens, tokens.indexOf(parenToken));
                    if (options.insertSpaceBeforeFunctionParenthesis) {
                        //insert a whitespace token
                        tokens.splice(tokens.indexOf(parenToken), 0, {
                            tokenType: brightscript_parser_1.TokenType.whitespace,
                            value: ' ',
                            startIndex: -1
                        });
                    }
                    //next loop iteration should be after the open paren
                    i = tokens.indexOf(parenToken);
                }
            }
            //empty curly braces
            if (token.tokenType === brightscript_parser_1.TokenType.openCurlyBraceSymbol && nextNonWhitespaceToken.tokenType === brightscript_parser_1.TokenType.closeCurlyBraceSymbol) {
                this.removeWhitespaceTokensBackwards(tokens, tokens.indexOf(nextNonWhitespaceToken));
                if (options.insertSpaceBetweenEmptyCurlyBraces) {
                    tokens.splice(tokens.indexOf(nextNonWhitespaceToken), 0, {
                        tokenType: brightscript_parser_1.TokenType.whitespace,
                        startIndex: -1,
                        value: ' '
                    });
                    //next loop iteration should be after the closing curly brace
                    i = tokens.indexOf(nextNonWhitespaceToken);
                }
            }
            //empty parenthesis (user doesn't have this option, we will always do this one)
            if (token.tokenType === brightscript_parser_1.TokenType.openParenSymbol && nextNonWhitespaceToken.tokenType === brightscript_parser_1.TokenType.closeParenSymbol) {
                this.removeWhitespaceTokensBackwards(tokens, tokens.indexOf(nextNonWhitespaceToken));
                //next loop iteration should be after the closing paren
                i = tokens.indexOf(nextNonWhitespaceToken);
            }
        }
        return tokens;
    };
    /**
     * Remove whitespace tokens backwards until a non-whitespace token is encountered
     * @param startIndex the index of the non-whitespace token to start with. This function will start iterating at `startIndex - 1`
     */
    BrightScriptFormatter.prototype.removeWhitespaceTokensBackwards = function (tokens, startIndex) {
        var removeCount = 0;
        var i = startIndex - 1;
        while (tokens[i--].tokenType === brightscript_parser_1.TokenType.whitespace) {
            removeCount++;
        }
        tokens.splice(startIndex - removeCount, removeCount);
    };
    /**
     * Remove whitespace until the next non-whitespace character.
     * This operates on the array itself
     */
    BrightScriptFormatter.prototype.removeWhitespace = function (tokens, index) {
        while (tokens[index] && tokens[index].tokenType === brightscript_parser_1.TokenType.whitespace) {
            tokens.splice(index, 1);
            index++;
        }
    };
    /**
     * Determine if the current token appears to be the negative sign for a numeric leteral
     */
    BrightScriptFormatter.prototype.looksLikeNegativeNumericLiteral = function (tokens, index) {
        var thisToken = tokens[index];
        if (thisToken.tokenType === brightscript_parser_1.TokenType.minusSymbol) {
            var nextToken = this.getNextNonWhitespaceToken(tokens, index);
            var previousToken = this.getPreviousNonWhitespaceToken(tokens, index);
            if (
            //next non-whitespace token is a numeric literal
            nextToken && nextToken.tokenType === brightscript_parser_1.TokenType.numberLiteral &&
                previousToken && BrightScriptFormatter.tokensBeforeNegativeNumericLiteral.indexOf(previousToken.tokenType) > -1) {
                return true;
            }
        }
        return false;
    };
    /**
     * Get the first token after the index that is NOT whitespace
     */
    BrightScriptFormatter.prototype.getNextNonWhitespaceToken = function (tokens, index) {
        for (index = index + 1; index < tokens.length; index++) {
            if (tokens[index] && tokens[index].tokenType !== brightscript_parser_1.TokenType.whitespace) {
                return tokens[index];
            }
        }
        //if we got here, we ran out of tokens. Return the EOF token
        return tokens[tokens.length - 1];
    };
    /**
     * Get the first token before the index that is NOT whitespace
     */
    BrightScriptFormatter.prototype.getPreviousNonWhitespaceToken = function (tokens, startIndex) {
        for (var i = startIndex - 1; i > -1; i--) {
            if (tokens[i] && tokens[i].tokenType !== brightscript_parser_1.TokenType.whitespace) {
                return tokens[i];
            }
        }
    };
    /**
     * Remove all trailing whitespace
     */
    BrightScriptFormatter.prototype.formatTrailingWhiteSpace = function (tokens, options) {
        var nextLineStartTokenIndex = 0;
        //the list of output tokens
        var outputTokens = [];
        //set the loop to run for a max of double the number of tokens we found so we don't end up with an infinite loop
        for (var outerLoopCounter = 0; outerLoopCounter <= tokens.length * 2; outerLoopCounter++) {
            var lineObj = this.getLineTokens(nextLineStartTokenIndex, tokens);
            nextLineStartTokenIndex = lineObj.stopIndex + 1;
            var lineTokens = lineObj.tokens;
            //the last token is newline or EOF, so the next-to-last token is where the trailing whitespace would reside
            var potentialWhitespaceTokenIndex = lineTokens.length - 2;
            var whitespaceTokenCandidate = lineTokens[potentialWhitespaceTokenIndex];
            //empty lines won't have any tokens
            if (whitespaceTokenCandidate) {
                //if the final token is whitespace, throw it away
                if (whitespaceTokenCandidate.tokenType === brightscript_parser_1.TokenType.whitespace) {
                    lineTokens.splice(potentialWhitespaceTokenIndex, 1);
                    //if the final token is a comment, trim the whitespace from the righthand side
                }
                else if (whitespaceTokenCandidate.tokenType === brightscript_parser_1.TokenType.quoteComment ||
                    whitespaceTokenCandidate.tokenType === brightscript_parser_1.TokenType.remComment) {
                    whitespaceTokenCandidate.value = trimRight(whitespaceTokenCandidate.value);
                }
            }
            //add this line to the output
            outputTokens.push.apply(outputTokens, lineTokens);
            //if we have found the end of file, quit the loop
            if (lineTokens[lineTokens.length - 1].tokenType === brightscript_parser_1.TokenType.END_OF_FILE) {
                break;
            }
        }
        return outputTokens;
    };
    BrightScriptFormatter.prototype.tokenIndexOf = function (tokenType, tokens) {
        for (var i = 0; i < tokens.length; i++) {
            var token = tokens[i];
            if (token.tokenType === tokenType) {
                return i;
            }
        }
        return -1;
    };
    /**
     * Get the tokens for the whole line starting at the given index (including the newline or EOF token at the end)
     * @param startIndex
     * @param tokens
     */
    BrightScriptFormatter.prototype.getLineTokens = function (startIndex, tokens) {
        var outputTokens = [];
        var index = startIndex;
        for (index = startIndex; index < tokens.length; index++) {
            var token = tokens[index];
            outputTokens[outputTokens.length] = token;
            if (token.tokenType === brightscript_parser_1.TokenType.newline ||
                token.tokenType === brightscript_parser_1.TokenType.END_OF_FILE) {
                break;
            }
        }
        return {
            startIndex: startIndex,
            stopIndex: index,
            tokens: outputTokens
        };
    };
    BrightScriptFormatter.prototype.normalizeOptions = function (options) {
        var fullOptions = {
            indentStyle: 'spaces',
            indentSpaceCount: BrightScriptFormatter.DEFAULT_INDENT_SPACE_COUNT,
            formatIndent: true,
            keywordCase: 'lower',
            compositeKeywords: 'split',
            removeTrailingWhiteSpace: true,
            keywordCaseOverride: {},
            formatInteriorWhitespace: true,
            insertSpaceBeforeFunctionParenthesis: false,
            insertSpaceBetweenEmptyCurlyBraces: false
        };
        if (options) {
            for (var attrname in options) {
                fullOptions[attrname] = options[attrname];
            }
        }
        if (!fullOptions.typeCase) {
            fullOptions.typeCase = fullOptions.keywordCase;
        }
        return fullOptions;
    };
    BrightScriptFormatter.prototype.isSingleLineIfStatement = function (lineTokens, allTokens) {
        var ifIndex = this.tokenIndexOf(brightscript_parser_1.TokenType.if, lineTokens);
        if (ifIndex === -1) {
            return false;
        }
        var thenIndex = this.tokenIndexOf(brightscript_parser_1.TokenType.then, lineTokens);
        var elseIndex = this.tokenIndexOf(brightscript_parser_1.TokenType.else, lineTokens);
        //if there's an else on this line, assume this is a one-line if statement
        if (elseIndex > -1) {
            return true;
        }
        //if there's no then, then it can't be a one line statement
        if (thenIndex === -1) {
            return false;
        }
        //if there's a comment at the end, this is a multi-line if statement
        if (this.tokenIndexOf(brightscript_parser_1.TokenType.remComment, lineTokens) > -1 || this.tokenIndexOf(brightscript_parser_1.TokenType.quoteComment, lineTokens) > -1) {
            return false;
        }
        //see if there is anything after the "then". If so, assume it's a one-line if statement
        for (var i = thenIndex + 1; i < lineTokens.length; i++) {
            var token = lineTokens[i];
            if (token.tokenType === brightscript_parser_1.TokenType.whitespace ||
                token.tokenType === brightscript_parser_1.TokenType.newline) {
                //do nothing with whitespace and newlines
            }
            else {
                //we encountered a non whitespace and non newline token, so this line must be a single-line if statement
                return true;
            }
        }
        return false;
    };
    /**
     * The default number of spaces when indenting with spaces
     */
    BrightScriptFormatter.DEFAULT_INDENT_SPACE_COUNT = 4;
    /**
     * Anytime one of these tokens are found before a minus sign,
     * we can safely assume the minus sign is associated with a negative numeric literal
     */
    BrightScriptFormatter.tokensBeforeNegativeNumericLiteral = [
        brightscript_parser_1.TokenType.plusSymbol,
        brightscript_parser_1.TokenType.minusSymbol,
        brightscript_parser_1.TokenType.asteriskSymbol,
        brightscript_parser_1.TokenType.forwardSlashSymbol,
        brightscript_parser_1.TokenType.backSlashSymbol,
        brightscript_parser_1.TokenType.additionAssignmentSymbol,
        brightscript_parser_1.TokenType.divisionAssignmentSymbol,
        brightscript_parser_1.TokenType.subtractionAssignmentSymbol,
        brightscript_parser_1.TokenType.multiplicationAssignmentSymbol,
        brightscript_parser_1.TokenType.integerDivisionAssignmentSymbol,
        brightscript_parser_1.TokenType.equalSymbol,
        brightscript_parser_1.TokenType.notEqual,
        brightscript_parser_1.TokenType.greaterThanSymbol,
        brightscript_parser_1.TokenType.greaterThanOrEqual,
        brightscript_parser_1.TokenType.lessThanSymbol,
        brightscript_parser_1.TokenType.lessThanOrEqual,
        brightscript_parser_1.TokenType.lessThanLessThanEqualSymbol,
        brightscript_parser_1.TokenType.greaterThanGreaterThanEqualSymbol,
        brightscript_parser_1.TokenType.return,
        brightscript_parser_1.TokenType.to,
        brightscript_parser_1.TokenType.step,
        brightscript_parser_1.TokenType.colonSymbol,
        brightscript_parser_1.TokenType.semicolonSymbol
    ];
    return BrightScriptFormatter;
}());
exports.BrightScriptFormatter = BrightScriptFormatter;
